// [[0], [1], [2], [3], [4], [5, 7, 8, 9, 10, 11], [6]]

#include <stdio.h>
#include "stdint.h"
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>

#include <list>
#include <vector>
#include <set>
#include <unordered_set>

extern "C" {
#include "csearch.h"
}


#define FLAG_MODE_P         1
#define FLAG_MODE_W         2
#define FLAG_MODE_ERROR     4
#define FLAG_MERGE          8
#define FLAG_APPROXIMATE    16

struct ELA2_Back_Edge {
    ELA2_Back_Edge(unsigned _header, unsigned _edge_weight) :
            header(_header), edge_weight(_edge_weight), 
            cross_region(false) {};
	ELA2_Back_Edge(const ELA2_Back_Edge & b) {
		header = b.header;
		edge_weight = b.edge_weight;
		cross_region = b.cross_region;
	}

    unsigned header, edge_weight;
    bool cross_region;
};

class Lukes_Partition {
public:
    Lukes_Partition(unsigned _flags, unsigned _fixed_cost) :
                child_1(0), child_2(0), subtree_root(0), flags(_flags),
                fixed_cost(_fixed_cost), first_weight(0), users(1),
                a_value(0), b_value(0), k_value(0) {};
    ~Lukes_Partition();

private:
    Lukes_Partition * child_1;
    Lukes_Partition * child_2;
    unsigned        subtree_root, flags, fixed_cost;
    unsigned        first_weight, users;
    uint64_t        a_value, b_value, k_value;

    std::vector<ELA2_Back_Edge> headers_reached;
    std::vector<ELA2_Back_Edge> headers_open;
    std::vector<ELA2_Back_Edge> headers_closed;

public:
    void add_first(unsigned first_member, Problem * problem);
    void add_footer(unsigned header, unsigned edge_weight);
    Lukes_Partition * merge_first_region(Lukes_Partition * part2);
    Lukes_Partition * concatenate(Lukes_Partition * part2);
    unsigned get_weight() const { return first_weight; };
    uint64_t get_gradient() const { return b_value; };
    uint64_t get_value_lower_bound(void);
    uint64_t get_value_for(unsigned weight);
    std::list<std::list<unsigned> > extract();
    void print ();
    bool is_greater(Lukes_Partition * other, unsigned expanded_weight);
    void expected_to_reached(Lukes_Partition * part2, bool cross_region);
    void increment(void) { users++; };
    bool decrement(void);
    bool verify(std::unordered_set<unsigned> & loop_antilink);
};

static void vextend(std::vector<ELA2_Back_Edge> & target,
            std::vector<ELA2_Back_Edge> & source)
{
    std::vector<ELA2_Back_Edge>::iterator fi;
    
    for (fi = source.begin(); fi != source.end(); fi++) {
        target.push_back((* fi));
    }
}
                
static void decdelete(Lukes_Partition * p)
{
    if (p && p->decrement()) {
        delete p;
    }
}

Lukes_Partition::~Lukes_Partition()
{
    decdelete(child_1);
    decdelete(child_2);
}

void Lukes_Partition::add_first(unsigned first_member, Problem * problem) 
{
    subtree_root = first_member;
    first_weight = problem->nodes[first_member].size;
    k_value = 0;
    b_value = 0;
    a_value = 0;
    if (first_member == 0) {
        a_value = 1;
    } else {
        a_value = problem->nodes[first_member].edge_weight;
    }
    // value = k_value + ((a_value + b_value) * weight)
    // a_value is total top-down edge weight for entry to the first region
    // b_value is total bottom-up edge weight for edges from that region
    // to some child region
}

Lukes_Partition * Lukes_Partition::merge_first_region(Lukes_Partition * part2) 
{
    Lukes_Partition * part1 = this;
    Lukes_Partition * target = new Lukes_Partition(flags, fixed_cost);

    // part2 is always a child region for part1.subtree_root
    // part2 is always a complete subtree
    //assert(problem->nodes[part2->subtree_root].parent == part1->subtree_root);

    part1->increment();
    part2->increment();

    target->subtree_root = part1->subtree_root;
    target->child_1 = part1;
    target->child_2 = part2;
    target->flags |= FLAG_MERGE;
    target->first_weight = part1->first_weight + part2->first_weight;

    target->a_value = part1->a_value;
    target->b_value = part1->b_value + part2->b_value;
    target->k_value = part1->k_value + part2->k_value;

    if (flags & FLAG_MODE_W) {
        // Headers expected & reached by part1 remain expected/reached
        // (Ancestor property - no new headers could have been found)
        vextend(target->headers_open, part1->headers_open);
        vextend(target->headers_closed, part1->headers_closed);
        vextend(target->headers_reached, part1->headers_reached);

        // Headers reached by part2 remain reached
        vextend(target->headers_reached, part2->headers_reached);

        // Headers expected by part2 must be checked to see if they
        // are now reached, they will not necessarily be cross-region
        target->expected_to_reached(part2, false);
    }
    return target;
}

Lukes_Partition * Lukes_Partition::concatenate(Lukes_Partition * part2) 
{
    Lukes_Partition * part1 = this;
    Lukes_Partition * target = new Lukes_Partition(flags, fixed_cost);

    // part2 is always a child region for part1.subtree_root
    // part2 is always a complete subtree
    //assert(problem->nodes[part2->subtree_root].parent == part1->subtree_root);

    part1->increment();
    part2->increment();

    target->subtree_root = part1->subtree_root;
    target->child_1 = part1;
    target->child_2 = part2;
    target->flags &= ~FLAG_MERGE;
    target->first_weight = part1->first_weight;

    target->a_value = part1->a_value;
    target->b_value = part1->b_value + part2->a_value;
    target->k_value = part1->k_value + part2->k_value;

    if (flags & FLAG_MODE_W) {
        // Headers expected & reached by part1 remain expected/reached
        // (Ancestor property - no new headers could have been found)
        vextend(target->headers_open, part1->headers_open);
        vextend(target->headers_closed, part1->headers_closed);
        vextend(target->headers_reached, part1->headers_reached);

        // Headers expected by part2 must be checked to see if they
        // are now reached, knowing that they will be cross-region now
        target->expected_to_reached(part2, true);

        // Headers reached by part2 are now final
        std::vector<ELA2_Back_Edge>::iterator fi;
        for (fi = part2->headers_reached.begin();
                fi != part2->headers_reached.end(); fi++) {
            if ((* fi).cross_region) {
                target->k_value += (uint64_t) (* fi).edge_weight * 
                        (uint64_t) (part2->first_weight + fixed_cost);
            }
        }
    }

    // part2's first region becomes frozen in place
    if (flags & FLAG_MODE_P) {
        target->k_value += part2->b_value;
    } else {
        if (!(flags & FLAG_MODE_W)) {
            target->k_value += part2->b_value * 
                        (part2->first_weight + fixed_cost);
        }
		target->k_value += part2->a_value * 
                        (part2->first_weight + fixed_cost);
    }
    return target;
}

void Lukes_Partition::expected_to_reached(Lukes_Partition * part2, 
                        bool cross_region)
{
    std::vector<ELA2_Back_Edge>::iterator fi;

    for (fi = part2->headers_open.begin();
            fi != part2->headers_open.end(); fi++) {
        ELA2_Back_Edge b = ELA2_Back_Edge((* fi));
	
        assert(!b.cross_region);
        if (cross_region) {
            b.cross_region = true;
        }
        assert(b.header <= subtree_root); 
        if (b.header == subtree_root) {
            headers_reached.push_back(b);
        } else if (b.cross_region) {
            headers_closed.push_back(b);
        } else {
            headers_open.push_back(b);
        }
    }
    for (fi = part2->headers_closed.begin();
            fi != part2->headers_closed.end(); fi++) {

        ELA2_Back_Edge b = ELA2_Back_Edge((* fi));
        assert(b.cross_region);
        assert(b.header <= subtree_root); 
        if (b.header == subtree_root) {
            headers_reached.push_back(b);
        } else {
            headers_closed.push_back(b);
        }
    }
}

uint64_t Lukes_Partition::get_value_lower_bound(void) 
{
    return get_value_for(first_weight);
}

uint64_t Lukes_Partition::get_value_for(unsigned weight) 
{
    if (flags & FLAG_MODE_P) {
        return k_value + ((a_value + b_value));
    }
    uint64_t total = k_value;
	
	total += ((uint64_t) (first_weight + fixed_cost) * (uint64_t) a_value);
	
    if (flags & FLAG_MODE_W) {
        // Assume headers reached are final
        std::vector<ELA2_Back_Edge>::iterator fi;
        for (fi = headers_reached.begin();
                fi != headers_reached.end(); fi++) {
            if ((* fi).cross_region) {
                total += (uint64_t) (* fi).edge_weight * 
                        (uint64_t) (weight + fixed_cost);
            }
        }
    } else {
        total += ((uint64_t) (weight + fixed_cost) * (uint64_t) b_value);
    }
    return total;
}

void Lukes_Partition::add_footer(unsigned header, unsigned edge_weight)
{
    headers_open.push_back(ELA2_Back_Edge(header, edge_weight));
}

void Lukes_Partition::print()
{
    std::list<std::list<unsigned> > p = extract();
    std::list<std::list<unsigned> >::iterator i;
    std::list<unsigned>::iterator j;
    
    printf("[");
    for (i = p.begin(); i != p.end(); i++) {
        printf("[");
        for (j = (* i).begin(); j != (* i).end(); j++) {
            printf("%u, ", (* j));
        }
        printf("], ");
    }
    printf("] %u %u %u\n",
        (unsigned) headers_reached.size(),
        (unsigned) headers_open.size(),
        (unsigned) headers_closed.size());
}

std::list<std::list<unsigned> > Lukes_Partition::extract() 
{
    // A slow operation to be used only rarely
    // Builds the complete partition out of the data structure
    // that has been formed by successive partition merges/concats


    if (!child_1) {
        std::list<unsigned> l1;
        std::list<std::list<unsigned> > l2;
        l1.push_back(subtree_root);
        l2.push_back(l1);
        return l2;
    }

    std::list<std::list<unsigned> > part1;
    std::list<std::list<unsigned> > part2;

    part1 = child_1->extract();
    part2 = child_2->extract();

    if (flags & FLAG_MERGE) {
        std::list<unsigned> r1;
        std::list<unsigned> r2;
        r1 = part1.front();
        r2 = part2.front();
        part1.pop_front();
        part2.pop_front();
        r1.splice(r1.end(), r2); // r1 += r2
        part1.push_front(r1);
    }
    part1.splice(part1.end(), part2); // part1 += part2
    return part1;
}

bool Lukes_Partition::verify(std::unordered_set<unsigned> & loop_antilink)
{
    if (!(flags & FLAG_MODE_W)) {
        return true;
    } else if (!child_1) {
        // not present
        return (loop_antilink.count(subtree_root) == 0);
    } else if (flags & FLAG_MERGE) {
        return child_1->verify(loop_antilink) 
                || child_2->verify(loop_antilink);
    } else {
        return child_1->verify(loop_antilink);
    }
}

bool Lukes_Partition::is_greater(Lukes_Partition * other,
                            unsigned expanded_weight)
{
    uint64_t v0 = this->get_value_for(expanded_weight);
    uint64_t v1 = other->get_value_for(expanded_weight);

    if (v0 > v1) {
        return true;
    } else if (v0 == v1) {
        return false; // (this->get_gradient() > other->get_gradient());
    } else {
        return false;
    }
}

bool Lukes_Partition::decrement(void)
{
    assert(users != 0);
    users--;
    return (users == 0);
}

class Lukes_Partition_Group {
public:
    Lukes_Partition_Group(Lukes_Partition * _initial_partition, Problem * _p)
    {
        problem = _p;
        weight = _initial_partition->get_weight();
		quantum_spm_size = (problem->spm_size / problem->quantum_size) + 1;
		quantum_size = problem->quantum_size;
		partition = new Lukes_Partition * [quantum_spm_size];
		memset(partition, 0, sizeof(Lukes_Partition *) * quantum_spm_size);
		merge_into_group(_initial_partition, weight);
	}
    
    ~Lukes_Partition_Group();

private:
    unsigned        weight, quantum_spm_size, quantum_size;
	Lukes_Partition ** partition;
    Problem *       problem;

public:
    void merge_into_group(
                    Lukes_Partition * new_partition, 
                    unsigned expanded_weight);
    Lukes_Partition * get(unsigned new_weight);
    void funk();
};

void Lukes_Partition_Group::funk()
{
#ifdef ENABLE_FUNK
    unsigned i;
    std::set<unsigned> collector;

    for (i = 1; i < quantum_spm_size; i++) {
        unsigned x = (unsigned) ((intptr_t) partition[i]);
        if (x && !collector.count(x)) {
            collector.insert(x); 
        }
    }

    static unsigned funkiest = 0;
    if (collector.size() > funkiest) {
        funkiest = collector.size();
        FILE * funky = fopen("funky.txt", "wt");
        assert(funky);
        fprintf(funky, "%u\n", funkiest);
        for (i = 1; i < quantum_spm_size; i++) {
            unsigned x = (unsigned) ((intptr_t) partition[i]);
            if (x && collector.count(x)) {
                collector.erase(x); 
                std::list<std::list<unsigned> > raw_part = 
                                        partition[i]->extract();
                std::list<std::list<unsigned> >::iterator pi;
                std::list<unsigned>::iterator ri;

                fprintf(funky, "@%u [", i);
                for (pi = raw_part.begin(); pi != raw_part.end(); pi++) {
                    fprintf(funky, "[");
                    for (ri = (* pi).begin(); ri != (* pi).end(); ri++) {
                        unsigned member = (* ri);

                        fprintf(funky, "%u, ", member);
                    }
                    fprintf(funky, "], ");
                }
                fprintf(funky, "]\n");
                fprintf(funky, "... %u %1.0f %1.0f\n",
                            partition[i]->get_weight(),
                            (double) partition[i]->get_gradient(),
                            (double) partition[i]->get_value_lower_bound());
            }
        }
        fprintf(funky, "\n\n");
        fclose(funky);
    }
#endif
}

void Lukes_Partition_Group::merge_into_group(
                    Lukes_Partition * new_partition, 
                    unsigned expanded_weight) 
{
    assert (weight == new_partition->get_weight());
    assert (weight <= expanded_weight);
	unsigned i = expanded_weight / quantum_size;
	assert (i < quantum_spm_size);
	Lukes_Partition * old_partition = partition[i];

    if (old_partition) {
        // old partition to compare with
        if (old_partition->is_greater(new_partition, expanded_weight)) {
            // replace
            decdelete(old_partition);
            partition[i] = new_partition;
            new_partition->increment();
        }
    } else {
        // new partition
        partition[i] = new_partition;
        new_partition->increment();
    }
}

Lukes_Partition * Lukes_Partition_Group::get(unsigned new_weight) 
{
    if (new_weight == 0) {
        new_weight = weight;
    }
	unsigned i = new_weight / quantum_size;
	assert(i < quantum_spm_size);
    Lukes_Partition * p = partition[i];
    assert(p);
    p->increment();
    return p;
}

Lukes_Partition_Group::~Lukes_Partition_Group()
{
	unsigned i;
	Lukes_Partition * p;
	for (i = 0; i < quantum_spm_size; i++) {
		p = partition[i];
        decdelete(p);
    }
	delete[] partition;
}


class Lukes_Fast_Method { 
public:
    Lukes_Fast_Method(unsigned _number, unsigned _flags,
                    Problem * _problem) {
        problem = _problem;
        flags = _flags;
        number = _number;
        size = problem->nodes[number].size;
		quantum_size = problem->quantum_size;
		quantum_spm_size = (problem->spm_size / quantum_size) + 1;
		now = new Lukes_Partition_Group * [quantum_spm_size];
		next = new Lukes_Partition_Group * [quantum_spm_size];
		memset(now, 0, sizeof(Lukes_Partition_Group *) * quantum_spm_size);
		memset(next, 0, sizeof(Lukes_Partition_Group *) * quantum_spm_size);
		fast = 0;
		total_non_zero = 0;
	};

    ~Lukes_Fast_Method();
   
    std::vector<unsigned> children;
    std::unordered_set<unsigned> loop_antilink;
    std::unordered_set<unsigned> must_merge;

private:
    Problem *       problem;
    unsigned        number, size, total_non_zero, flags;
	unsigned		quantum_size, quantum_spm_size;
    Lukes_Partition_Group ** now;
    Lukes_Partition_Group ** next;
	unsigned *		fast;

public:
    void swap_holders();
    void clear();
    void merge_partition(unsigned new_weight, 
                unsigned expanded_weight, Lukes_Partition * new_partition);
    void set_partition(Lukes_Partition * partition);
    Lukes_Partition * get_partition();
    Lukes_Partition * pick_best(unsigned actual_weight, 
                unsigned new_weight);
    void operation(Lukes_Fast_Method * child);
    bool finalise();
};

Lukes_Fast_Method::~Lukes_Fast_Method() 
{
    clear();
	delete[] now;
	delete[] next;
}

void Lukes_Fast_Method::clear()
{
    swap_holders();
    swap_holders();
	if (fast) {
		delete[] fast;
	}
    fast = 0;
    total_non_zero = 0;
    loop_antilink.clear();
    must_merge.clear();
}

void Lukes_Fast_Method::swap_holders()
{
	unsigned i;
	for (i = 0; i < quantum_spm_size; i++) {
		delete now[i];
		now[i] = next[i];
		next[i] = 0;
	}
}

void Lukes_Fast_Method::merge_partition(unsigned new_weight, 
                unsigned expanded_weight, Lukes_Partition * new_partition)
{
	unsigned i = new_weight / quantum_size;
	assert(i < quantum_spm_size);
	Lukes_Partition_Group * pg = next[i];
    if (pg) {
        pg->merge_into_group(new_partition, expanded_weight);
    } else {
        pg = new Lukes_Partition_Group(new_partition, problem);
        next[i] = pg;

        if (new_weight != expanded_weight) {
            pg->merge_into_group(new_partition, expanded_weight);
        }
    }
}

void Lukes_Fast_Method::set_partition(Lukes_Partition * partition)
{
    assert(!now[0]);
	now[0] = new Lukes_Partition_Group(partition, problem);
}

Lukes_Partition * Lukes_Fast_Method::get_partition()
{
    return pick_best(0, 0);
}

Lukes_Partition * Lukes_Fast_Method::pick_best(unsigned actual_weight,
                    unsigned new_weight)
{
	unsigned i = actual_weight / quantum_size;
	assert(i < quantum_spm_size);
    
	Lukes_Partition_Group * p = now[i];
	assert(p);
    return p->get(new_weight);
}

void Lukes_Fast_Method::operation(Lukes_Fast_Method * child)
{
    unsigned pi, fi, fin;
	Lukes_Partition * part1;
    Lukes_Partition * part2;
    Lukes_Partition * previous1;
    Lukes_Partition * previous2;
    Lukes_Partition * previous12;
    unsigned expansion;

    for (pi = 0; pi < quantum_spm_size; pi++) {
        Lukes_Partition_Group * parent_group = now[pi];
		if (!parent_group) continue;

		unsigned parent_weight = pi * quantum_size;
        unsigned max_expansion = problem->spm_size + 1 - parent_weight;
        bool previous12ok;

        previous12ok = false;
        previous1 = previous12 = 0;
        for (expansion = 0; 
                (!must_merge.count(child->number))
                && (expansion < max_expansion); 
                    expansion += problem->quantum_size) {

            part1 = parent_group->get(
                    ((flags & FLAG_MODE_ERROR) ? 0 : expansion) + 
                            parent_weight);
            if (part1 != previous1) {
                decdelete(previous1);
                decdelete(previous12);

                previous1 = part1;
                previous1->increment();

                part2 = child->pick_best(0, 0);
                previous12 = part1->concatenate(part2);
                decdelete(part2);

                previous12ok = part1->verify(loop_antilink);
            }
            decdelete(part1);

            if (previous12ok) {
                merge_partition(parent_weight,
                        expansion + parent_weight, previous12);
            }
        }
        decdelete(previous1);
        decdelete(previous12);

        assert(child->fast);
        fin = child->total_non_zero;
        previous1 = 0;
        previous2 = 0;
        previous12 = 0;

        for (fi = 0; fi < fin; fi++) {
            unsigned ci = child->fast[fi];
            Lukes_Partition_Group * child_group = child->now[ci];
            unsigned child_weight = ci * quantum_size;
            unsigned new_weight = child_weight + parent_weight;
            assert(child_group);

            if (new_weight > problem->spm_size) break;

            for (expansion = 0; 
                    expansion < max_expansion; 
                    expansion += problem->quantum_size) {

                if ((new_weight + expansion) > problem->spm_size) break;

                part1 = parent_group->get(
                    ((flags & FLAG_MODE_ERROR) ? 0 : expansion) + new_weight);
                part2 = child_group->get(
                    ((flags & FLAG_MODE_ERROR) ? 0 : expansion) + new_weight);

                if ((part1 != previous1) || (part2 != previous2)) {
                    decdelete(previous1);
                    decdelete(previous2);
                    decdelete(previous12);
                    previous1 = part1;
                    previous2 = part2;
                    previous1->increment();
                    previous2->increment();

                    previous12 = part1->merge_first_region(part2);
                    previous12ok = previous12->verify(loop_antilink);
                }
                if (previous12ok) {
                    merge_partition(new_weight,
                            expansion + new_weight, previous12);
                }

                decdelete(part1);
                decdelete(part2);
            }
        }
        decdelete(previous1);
        decdelete(previous2);
        decdelete(previous12);
    }
    swap_holders();
}

bool Lukes_Fast_Method::finalise()
{
    Lukes_Partition * best_part = 0;
    uint64_t best_value = 0;
	unsigned pi, fi;

	assert(!fast);
	assert(!total_non_zero);
    for (pi = 0; pi < quantum_spm_size; pi++) {
        Lukes_Partition_Group * g = now[pi];

		if (!g) continue;

		unsigned new_weight = pi * quantum_size;
        Lukes_Partition * part = g->get(new_weight);
        uint64_t value = part->get_value_lower_bound();

        if ((!best_part) || (best_value > value)) {
            decdelete(best_part);
            best_part = part;
            best_value = value;
        } else {
            decdelete(part);
        }
		assert(pi != 0);
		total_non_zero++;
        g->funk();
	}
    if (total_non_zero == 0) {
        // No valid partitions; stop!
        return false;
    }

	assert(total_non_zero);
	fast = new unsigned[total_non_zero];
	for (pi = fi = 0; pi < quantum_spm_size; pi++) {
        Lukes_Partition_Group * g = now[pi];

		if (g) {
			fast[fi] = pi;
			fi++;
		}
    }
	assert(fi == total_non_zero);
    assert(best_part);
    set_partition(best_part);
    decdelete(best_part);
    return true;
}

Lukes_Partition * algo_core(
            std::vector<Lukes_Fast_Method *> & master_list,
            unsigned flags, Problem * p, uint64_t be_link)
{
    unsigned i, j;
    unsigned expanded_weight;

    for (i = 0; i < p->num_nodes; i++) {
		Lukes_Fast_Method * parent = master_list[i];
        parent->clear();
    }
    for (i = 0; i < p->num_nodes; i++) {
        Lukes_Partition * part = new Lukes_Partition(flags, p->fixed_cost);

        part->add_first(i, p);
        for (j = 0; j < p->num_back_edges; j++) {
            Back_Edge * be = &p->back_edges[j];

            if (be->footer == i) {
                part->add_footer(be->header, be->edge_weight);
            }
        }

        for (expanded_weight = p->nodes[i].size;
                expanded_weight <= p->spm_size; 
                expanded_weight += p->quantum_size) {
            master_list[i]->merge_partition(
                    p->nodes[i].size, expanded_weight, part);
        }
        master_list[i]->swap_holders();
        p->nodes[i].region = UINT_MAX;
        decdelete(part);
    }

    if (flags & FLAG_MODE_W) {
        unsigned valid_back_edge_count = 0;
        for (j = 0; j < p->num_back_edges; j++) {
            Back_Edge * be = &p->back_edges[j];

            unsigned x, y = be->footer;
            unsigned size = 0;
            while ((y != 0) && (y != be->header)) {
                assert(y < p->num_nodes);
                size += p->nodes[y].size;
                x = p->nodes[y].parent;
                y = x;
            }
            assert(y == be->header); // ancestor check
            size += p->nodes[y].size;

            if (size > p->spm_size) {
                continue;
            }

            if ((be_link >> (uint64_t) valid_back_edge_count) & (uint64_t) 1) {
                y = be->footer;
                while ((y != 0) && (y != be->header)) {
                    assert(y < p->num_nodes);
                    x = p->nodes[y].parent;
                    master_list[x]->must_merge.insert(y);
                    y = x;
                }
                assert(y == be->header); // ancestor check
            } else {
                master_list[be->header]->loop_antilink.insert(be->footer);
            }
            valid_back_edge_count++;
        }
    }

    i = p->num_nodes;
    while (i != 0) {
        i--;

		std::vector<unsigned>::iterator ci;
		Lukes_Fast_Method * parent = master_list[i];
		Lukes_Fast_Method * child;
		
		for (ci = parent->children.begin(); 
                    ci != parent->children.end(); ci++) {
			child = master_list[(* ci)];
			parent->operation(child);
        }
        if (!parent->finalise()) {
            return 0;
        }
    }
    Lukes_Partition * part = master_list[0]->get_partition();
    return part;
}

extern "C" uint64_t extended_lukes(Problem * p);

uint64_t extended_lukes(Problem * p)
{
    std::vector<Lukes_Fast_Method *> master_list;
    unsigned i;
    unsigned flags = 0;

    switch(p->mode) {
    case 'P':   flags |= FLAG_MODE_P; 
                // Lukes' algorithm
                break;
    case 'w':   flags |= FLAG_APPROXIMATE;
    case 'W':   flags |= FLAG_MODE_W; 
                // ELA-2
                break;
    case 'E':   flags |= FLAG_MODE_ERROR; 
                // Error is Lukes' algorithm with \Gamma_2(R) substituted
                // ...gets it wrong
                break;
    case 'C':   flags |= 0; 
                // ELA-1
                break;
    default:    assert(false); 
                break;
    }

    for (i = 0; i < p->num_nodes; i++) {
        master_list.push_back(new Lukes_Fast_Method(i, flags, p));
    }

    for (i = 1; i < p->num_nodes; i++) {
        Node * cn = &p->nodes[i];
        master_list[cn->parent]->children.push_back(i);
    }

    uint64_t be_link = 0, be_num = 1;
    unsigned valid_back_edge_count = 0;

    if (flags & FLAG_MODE_W) {
        for (i = 0; i < p->num_back_edges; i++) {
            Back_Edge * be = &p->back_edges[i];

            unsigned pos = be->footer;
            unsigned size = 0;

            while ((pos != 0) && (pos != be->header)) {
                assert(pos < p->num_nodes);
                size += p->nodes[pos].size;
                pos = p->nodes[pos].parent;
            }
            assert(pos == be->header); // ancestor check
            size += p->nodes[pos].size;

            printf("size %u   %u\n", size, p->spm_size);
            if (size <= p->spm_size) {
                valid_back_edge_count ++;
            }
        }

        printf("%u valid back edges\n", valid_back_edge_count);
        if (flags & FLAG_APPROXIMATE) {
            assert(valid_back_edge_count < 64);
        } else {
            assert(valid_back_edge_count < 16);
        }
        be_num = (uint64_t) 1 << (uint64_t) valid_back_edge_count;
    }
  
    assert(be_num);

    uint64_t best_value = (uint64_t) -1;
    uint64_t value;
    Lukes_Partition * best_part = 0;
    Lukes_Partition * part = 0;

    if (flags & FLAG_APPROXIMATE) {
        uint64_t test = 0;

        best_part = algo_core(master_list, flags, p, 0);
        best_value = best_part->get_value_lower_bound();

        for (test = 1; test < be_num; test *= 2) {
            printf("exlukes: %ju of %ju\n", 
                    1 + test, be_num);
            part = algo_core(master_list, flags, p, be_link | test);
            if (part) {
                value = part->get_value_lower_bound();

                printf("value = %ju\n", value);
                if (value <= best_value) {
                    decdelete(best_part);
                    best_part = part;
                    best_value = value;
                    be_link |= test;
                } else {
                    decdelete(part);
                    be_link &= ~test;
                }
            } else {
                printf("value = undef\n");
                be_link &= ~test;
            }
        }
    } else {
        for (be_link = 0; be_link < be_num; be_link++) {
            printf("exlukes: %ju of %ju\n", 
                    1 + be_link, be_num);
            part = algo_core(master_list, flags, p, be_link);
            if (part) {
                value = part->get_value_lower_bound();

                printf("value = %ju\n", value);
                if (value < best_value) {
                    decdelete(best_part);
                    best_part = part;
                    best_value = value;
                } else {
                    decdelete(part);
                }
            } else {
                printf("value = undef\n");
            }
        }
    }

    assert(best_part);
    assert(best_value != ((uint64_t) -1));
    std::list<std::list<unsigned> > raw_part = best_part->extract();
    std::list<std::list<unsigned> >::iterator pi;
    std::list<unsigned>::iterator ri;

    i = 0;
    for (pi = raw_part.begin(); pi != raw_part.end(); pi++) {
        for (ri = (* pi).begin(); ri != (* pi).end(); ri++) {
            unsigned member = (* ri);

            p->nodes[member].region = i;
        }
        i++;
    }
    decdelete(best_part);
    assert(p->nodes[0].region == 0);
    for (i = 0; i < p->num_nodes; i++) {
        assert(p->nodes[i].region != UINT_MAX);
        delete master_list[i];
    }
    if (flags & FLAG_MODE_P) {
        best_value--;
    }

    return best_value;
}
   


